import os
import numpy as np
from ase import Atom, Atoms
from ase.data import chemical_symbols
from aegon.libposcar import tag
#------------------------------------------------------------------------------------------
def readcfgs(filename):
    if not os.path.isfile(filename):
        print(f"El archivo {filename} no existe.")
        return []
    moleculas = [] 
    singlemol = None  
    natom = 0 
    with open(filename, 'r') as contcarfile:
        for line in contcarfile:
            line = line.strip()
            if line == "BEGIN_CFG":
                if singlemol: 
                    moleculas.append(singlemol)
                singlemol = Atoms()  
            elif "Size" in line:
                size_line = contcarfile.readline().strip()
                natom = int(size_line)
            elif "AtomData" in line:
                for _ in range(natom):
                    coord_line = contcarfile.readline().strip()
                    ls = coord_line.split()
                    numero_atomico = int(5) #Escribir ls[1]  
                    ss = chemical_symbols[numero_atomico] 
                    xc, yc, zc = float(ls[2]), float(ls[3]), float(ls[4])
                    ai = Atom(symbol=ss, position=(xc, yc, zc))
                    singlemol.append(ai)
                    #print(f"tomo aadido: {ai}") 
            elif line == "Energy":
                energy_line = contcarfile.readline().strip()
                energy = float(energy_line)
                singlemol.info['e'] = energy
                singlemol.info['i'] = '_'
                #print(f"Energa registrada: {energy}")
        if singlemol:
            moleculas.append(singlemol)
    return moleculas
#------------------------------------------------------------------------------------------
def writecfgs(poscarlist, file, force=False):
    if not isinstance(poscarlist, list): poscarlist = [poscarlist]
    f=open(file,"w")
    poscarlist=tag(poscarlist)
    for atoms in poscarlist:
        print("BEGIN_CFG", file=f)
        print(" Size", file=f)
        natoms=len(atoms)
        print("    %d" %(natoms), file=f)
        if np.any(atoms.pbc):
            print(" Supercell", file=f)
            matrix=atoms.cell
            print("       %10.6f    %10.6f    %10.6f" %(matrix[0,0],matrix[0,1],matrix[0,2]), file=f)
            print("       %10.6f    %10.6f    %10.6f" %(matrix[1,0],matrix[1,1],matrix[1,2]), file=f)
            print("       %10.6f    %10.6f    %10.6f" %(matrix[2,0],matrix[2,1],matrix[2,2]), file=f)
        if force:
            print(" AtomData:  id type       cartes_x      cartes_y      cartes_z           fx          fy          fz",file=f)
            forces=atoms.arrays['forces']
            for k, atom in enumerate(atoms):
                xc, yc, zc = atom.position
                fx, fy, fz=forces[k]
                tipo = atom.tag
                print("            %2d    %d     %10.6f    %10.6f    %10.6f    %9.6f   %9.6f   %9.6f" %(k+1,tipo,xc,yc,zc,fx,fy,fz), file=f)
        else:
            print(" AtomData:  id type       cartes_x      cartes_y      cartes_z", file=f)
            for k, atom in enumerate(atoms):
                xc, yc, zc = atom.position
                tipo = atom.tag
                print("             %d    %d     %10.6f    %10.6f    %10.6f" %(k+1,tipo,xc,yc,zc), file=f)
        print(" Energy", file=f)
        energy=atoms.info['e']
        print("        %12.8f" %(energy), file=f)
        print(" Feature   EFS_by    VASP", file=f)
        print("END_CFG\n", file=f)
    f.close()
#------------------------------------------------------------------------------------------



import os
import numpy as np
from ase.data import covalent_radii
from ase.data.colors import cpk_colors
#------------------------------------------------------------------------------------------
aspect_ratios_4by3 = {
    'a': (800, 600),     #SVGA
    'b': (1024, 768),    #XGA
    'c': (1280, 960),    #UXGA
    'd': (1600, 1200),
    'e': (2048, 1536),
    'f': (3200, 2400),
    'g': (6000, 4500),
}
aspect_ratios_16by9 = {
    'a': (1280, 720),    # HD
    'b': (1920, 1080),   # Full HD
    'c': (2560, 1440),   # QHD/2K
    'd': (3840, 2160),   # 4K UHD
    'e': (7680, 4320),   # 8K UHD
}
#------------------------------------------------------------------------------------------
bwd='0.12'                    # 0.15 GRUESO DE LOS ENLACES
dlv=0.15                      # GRUESO DEL LADO DEL CUBO
sphere_factor=float(0.55)     # 0.5 (g 1.5)
projection='orthographic'     # 'perspective'
reflection='reflection 0.0'
reflection_model='phong 1.0'  # 'specular'
radio_factor=float(1.30)      # 1.4 1.6 #ALCANCE DE LOS ENLACES
tmit = 0.35                   # transmit (transparencia) 0.35 0.75
#-----------------------------------------------------------------------------------------
def write_image(poscarin, basename, f=1.2, quality='f'):
    if quality in aspect_ratios_4by3:
        width, height = aspect_ratios_4by3[quality]
        print(f"Width: {width}, Height: {height}")
    else:
        print(f"'{quality}' is not defined.")
    poscarxx=poscarin.copy()
    matrix=poscarxx.cell
    a1=np.array(matrix[0,:])
    a2=np.array(matrix[1,:])
    a3=np.array(matrix[2,:])
    a=np.linalg.norm(a1)
    b=np.linalg.norm(a2)
    c=np.linalg.norm(a3)
    d=max([a,b,c])
    factor=f*d
    background1,background2,background3='2.0','2.0','2.0'
    light11,light12,light13='1','1','1'
    light21,light22,light23='1','1','1'
    camara_rotate='y*0.0'
    camara_loca1,camara_loca2,camara_loca3='0','0',str(factor)
    camara_look1,camara_look2,camara_look3='0','0','0'
    name_pov=basename+'.pov' 
    name_png=basename+'.png'
    opnew = open(name_pov,'w')
    print('global_settings { assumed_gamma 1.0 }', file=opnew)
    #print('# include \'colors.inc\'', file=opnew)
    print('background{color rgb<2.0, 2.0, 2.0>}\n', file=opnew)
    print('light_source {< 10, -8, -8> color rgb <1, 1, 1>}', file=opnew)
    print('light_source {< -8,  8,  8> color rgb <1, 1, 1>}\n', file=opnew)
    print('camera {%s location <0, 0, %4.2f> look_at <0, 0, 0> rotate y*0.0}\n' %(projection, factor), file=opnew)
    def primo(v1,v2):
        print('cylinder {<%9.6f, %9.6f, %9.6f> <%9.6f, %9.6f, %9.6f>, %6.4f pigment{color rgb <0, 0, 0>} finish {%s %s}}' %(-v1[0],v1[1],v1[2],-v2[0],v2[1],v2[2],dlv,reflection_model,reflection), file=opnew)
    #PARA QUE LA FIGURA SIEMPRE SALGA CENTRADA
    a0=-(a1+a2+a3)/2
    primo(a0,a0+a1)
    primo(a0,a0+a2)
    primo(a0,a0+a3)
    primo(a0+a1,a0+a1+a2)
    primo(a0+a1,a0+a1+a3)
    primo(a0+a2,a0+a1+a2)
    primo(a0+a2,a0+a2+a3)
    primo(a0+a3,a0+a1+a3)
    primo(a0+a3,a0+a2+a3)
    primo(a0+a1+a2,a0+a1+a2+a3)
    primo(a0+a1+a3,a0+a1+a2+a3)
    primo(a0+a2+a3,a0+a1+a2+a3)
    nn=len(poscarxx)
    for ii in range(nn):
        ni=poscarxx[ii].number
        ri=covalent_radii[ni]
        xxi,yyi,zzi=np.array(poscarxx[ii].position) + a0
        for jj in range(ii+1,nn):
            nj=poscarxx[jj].number
            rj= covalent_radii[nj]
            xxj,yyj,zzj=np.array(poscarxx[jj].position) + a0
            uij=poscarxx[jj].position - poscarxx[ii].position
            rr=np.linalg.norm(uij)
            uijn=uij/rr
            rt=rr/(ri+rj)
            if rt < radio_factor:
                xpm, ypm, zpm=(xxj+xxi)/2.0, (yyj+yyi)/2.0, (zzj+zzi)/2.0
                cpki=cpk_colors[ni]
                cpkj=cpk_colors[nj]
                kolori=f"<{', '.join(f'{x:.3f}' for x in cpki)}>"
                kolorj=f"<{', '.join(f'{x:.3f}' for x in cpkj)}>"
                otf=float(0.95)
                xxic=xxi+sphere_factor*otf*ri*uijn[0]
                yyic=yyi+sphere_factor*otf*ri*uijn[1]
                zzic=zzi+sphere_factor*otf*ri*uijn[2]
                xxjc=xxj-sphere_factor*otf*rj*uijn[0]
                yyjc=yyj-sphere_factor*otf*rj*uijn[1]
                zzjc=zzj-sphere_factor*otf*rj*uijn[2]
                print('cylinder {<%9.6f, %9.6f, %9.6f> <%9.6f, %9.6f, %9.6f> %s pigment {color rgb %s} finish {%s %s}}' %(-xxic,yyic,zzic,-xpm,ypm,zpm,bwd,kolori,reflection_model,reflection), file=opnew)
                print('cylinder {<%9.6f, %9.6f, %9.6f> <%9.6f, %9.6f, %9.6f> %s pigment {color rgb %s} finish {%s %s}}' %(-xpm,ypm,zpm,-xxjc,yyjc,zzjc,bwd,kolorj,reflection_model,reflection), file=opnew)
    for iatom in poscarxx:
        cr=covalent_radii[iatom.number]
        radii_div2=sphere_factor*cr
        cpk=cpk_colors[iatom.number]
        kolor=f"<{', '.join(f'{x:.3f}' for x in cpk)}>"
        xx, yy, zz=np.array(iatom.position) + a0
        s0='pigment {color rgb'
        print('sphere {<%9.6f, %9.6f, %9.6f>, %6.4f pigment {color rgb %s} finish {%s %s}}' %(-xx,yy,zz,radii_div2,kolor,reflection_model,reflection), file=opnew)
    opnew.close()
    comp='povray +A Display=Off Output_File_Type=N All_Console=Off Width='+str(width)+' Height='+str(height)+' '+name_pov+' > /dev/null 2>&1'
    os.system(comp)
    print('Output= %s %s' %(name_pov, name_png))
    doma='rm -f '+str(name_pov)
    os.system(doma)
#-----------------------------------------------------------------------------------------



from joblib import Parallel, delayed
import numpy as np
from dscribe.descriptors import CoulombMatrix,  ACSF, SOAP, MBTR, LMBTR
from libdiscusr import USRMonoatom#, USRMultiatom
#------------------------------------------------------------------------------
def descriptors_MBTR_dis(lista):
    des_list = []
    for imol in lista:
        atoms = imol.get_chemical_symbols()
        especies = set(atoms)
        mbtr = MBTR(
            species=especies,
            geometry={"function": "distance"},
            grid={"min": 0, "max": 15, "sigma": 0.075, "n": 500},
            weighting={"function": "exp", "r_cut": 10, "threshold": 1e-3},
            periodic=False, sparse=False, normalization="l2"
        )
        des = mbtr.create(imol)
        des_list.append(des)
    return des_list

def descriptors_MBTR_angle(lista):
    des_list = []
    for imol in lista:
        atoms = imol.get_chemical_symbols()
        especies = set(atoms)
        mbtr = MBTR(
            species=especies,
            geometry={"function": "angle"},
            grid={"min": 0, "max": 180, "sigma": 1, "n": 360},
            weighting={"function": "exp", "r_cut": 10, "threshold": 1e-3},
            periodic=False, sparse=False, normalization="l2"
        )
        des = mbtr.create(imol)
        des_list.append(des)
    return des_list

def descriptors_MBTR_inv_dis (lista):
    des_list = []
    for  imol in lista:
        atoms = imol.get_chemical_symbols()
        especies = set(atoms)
        mbtr = MBTR(species = especies, geometry={"function": "inverse_distance"},
                    grid = {"min": 0, "max": 1, "sigma": 0.005, "n": 500}, 
                    weighting={"function": "exp", "r_cut": 10, "threshold": 1e-3}, #weighting={"function": "inverse_square", "r_cut": 10},
                    periodic=False, sparse=False, normalization="l2")
        des = mbtr.create(imol)
        des_list.append(des)
    return des_list

def descriptors_MBTR_cos (lista):
    des_list = []
    for  imol in lista:
        atoms = imol.get_chemical_symbols()
        especies = set(atoms)
        mbtr = MBTR(species = especies, geometry={"function": "cosine"},
                    grid = {"min": -1.0, "max": 1.0, "sigma": 0.01, "n": 300}, 
                    weighting={"function": "exp", "r_cut": 10, "threshold": 1e-3}, #weighting={"function": "inverse_square", "r_cut": 10},
                    periodic=False, sparse=False, normalization="l2")
        des = mbtr.create(imol)
        des_list.append(des)
    return des_list

def descriptors_LMBTR (lista):
    des_list = []
    for imol in lista:
         geometria={"function": "distance"}
         grid={"min": 0, "max": 5, "n": 100, "sigma": 0.1}
         pesaje={"function": "exp", "scale": 0.5, "threshold": 1e-3}
         atoms = imol.get_chemical_symbols()
         especies = set(atoms)
         lmbtr = LMBTR(species = especies, geometry=geometria, grid=grid, weighting=pesaje, periodic=False, normalization="l2")
         lmbtr_des = lmbtr.create(imol)
         des = lmbtr_des.flatten()
         des_list.append(des)
    return des_list

def descriptors_SOAP (lista):
    des_list = []
    for imol in lista:
        atoms = imol.get_chemical_symbols()
        especies = set(atoms)
        r_cut = 6.0
        n_max = 8
        l_max = 6
        soap = SOAP(species = especies, periodic=False, r_cut = r_cut, n_max=n_max, l_max=l_max)
        soap_des = soap.create(imol)
        des = soap_des.flatten()
        des_list.append(des)
    return des_list

def descriptors_CoulombMatrix (lista):
    des_list = []
    for imol in lista:
        filas = len(imol)
        cm = CoulombMatrix(n_atoms_max = filas)
        des = cm.create(imol)
        des_list.append(des)
    return des_list

def descriptors_ACSF (lista):
    des_list = []
    for imol in lista:
         g2params=[[1, 1], [1, 2], [1, 3]]
         g4params=[[1, 1, 1], [1, 2, 1], [1, 1, -1], [1, 2, -1]]
         atoms = imol.get_chemical_symbols()
         especies = set(atoms)
         r_cut = 3.5
         acsf = ACSF(species = especies, r_cut = r_cut, g2_params=g2params, g4_params=g4params)
         acsf_des = acsf.create(imol)
         des = acsf_des.flatten()
         des_list.append(des)
    return des_list

def descriptors_USR_mono (lista):
    des_list = []
    for imol in lista:
        des = USRMonoatom(imol)
        des_list.append(des)
    return des_list

#def descriptors_USR_multi(lista):
#    des_list = []
#    for imol in lista:
#        des = USRMultiatom(imol)
#        des_list.append(des)
#    return des_list
#-------------------------------------------------------------------------------

def split_list(lista, chunk_size):
    return [lista[i:i + chunk_size] for i in range(0, len(lista), chunk_size)]

def descriptors_sel(lista, des, chunk_size=500, n_jobs=-1):
    descriptor_functions = { "MBTR_dis": descriptors_MBTR_dis, "MBTR_ang": descriptors_MBTR_angle,
                             "MBTR_inv_dis": descriptors_MBTR_inv_dis, "MBTR_cos": descriptors_MBTR_cos,
                             "USR_mono": descriptors_USR_mono, #"USR_multi": descriptors_USR_multi,
                             "LMBTR": descriptors_LMBTR, "SOAP": descriptors_SOAP,
                             "CoulombMatrix": descriptors_CoulombMatrix, "ACSF": descriptors_ACSF}
    
    des_fun = descriptor_functions[des]
    sublistas = split_list(lista, chunk_size)
    resultados = Parallel(n_jobs=n_jobs)(delayed(des_fun)(sublista) for sublista in sublistas)
    descriptores = [descriptor for sublista in resultados for descriptor in sublista]
    return descriptores

def descriptors_comb(lista, des_1, des_2, chunk_size=500, n_jobs=-1): 
    des_list1 = descriptors_sel(lista, des_1, chunk_size, n_jobs)
    des_list2 = descriptors_sel(lista, des_2, chunk_size, n_jobs)
    combined = [np.concatenate((d1, d2)) for d1, d2 in zip(des_list1, des_list2)]
    return combined



import os
from aegon.libposcar import molecule2poscar
from aegon.libcfg import readcfgs, writecfgs
# -------------------------------------------------------------------
def make_mlp_relax_files(fileout, mlp_bin, mtp_pot):
    fh=open(fileout,'w')
    print("#!/bin/bash", file=fh)
    print("export OMP_NUM_THREADS=32", file=fh)
    print("MLP=%s" %(mlp_bin), file=fh)
    print("$MLP relax relax.ini --cfg-filename=initial.cfg --save-relaxed=relaxed.cfg --force-tolerance=1e-6 --stress-tolerance=1e-6", file=fh)
    print("if [ -f \"relaxed.cfg_0\" ]; then mv \"relaxed.cfg_0\" \"relaxed.cfg\"; fi", file=fh)
    fh.close()
    fg=open('relax.ini','w')
    print("mtp-filename           %s" %(mtp_pot), file=fg)
    fg.close()
# -------------------------------------------------------------------
def opt_mlip(molecules, mlp_bin, mtp_pot):
    poscars = molecule2poscar(molecules, 15.0)
    writecfgs(poscars,'initial.cfg', force=False)
    make_mlp_relax_files('relax.sh', mlp_bin, mtp_pot)    
    os.system('sh relax.sh')
    asemols=readcfgs('relaxed.cfg')
    os.system('rm -f initial.cfg relax.ini relax.sh')
    return asemols
# -------------------------------------------------------------------
